import argparse
from time import time_ns

import librosa
import soundfile as sf
from numpy import floating
from numpy.typing import NDArray

from .v2 import vsed_debug_v2


def cli_runner() -> None:
    # parse argument
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str)
    parser.add_argument("output", type=str)
    args = parser.parse_args()
    # trimming
    y, sr = librosa.load(args.input, sr=48000)
    y_trimmed: NDArray[floating] = trim(y, sr)
    sf.write(args.output, y_trimmed, sr, format="WAV", subtype="PCM_24")


def trim(y: NDArray[floating], sr: float) -> NDArray[floating]:
    s_sec, e_sec = vsed(y, sr)
    return y[int(s_sec * sr) : int(e_sec * sr)]


def vsed(
    y: NDArray[floating], sr: float, seed: int = time_ns() % 2**32
) -> tuple[float, float]:
    # shape check (monaural only)
    if y.ndim != 1:
        raise ValueError("PySaten only supports monaural audio.")
    # trim
    return vsed_debug_v2(y, sr, noise_seed=seed).get_result()



import numpy as np
from numpy import floating
from numpy.typing import NDArray


def root_mean_square(
    y: NDArray[floating], win_length: int, hop_length: int
) -> NDArray[floating]:
    rms = np.zeros(int(np.ceil(float(len(y)) / hop_length)))
    for i in range(len(rms)):
        # get target array
        idx = i * hop_length
        zc_start = int(max(0, idx - (win_length / 2)))
        zc_end = int(min(idx + (win_length / 2), len(y) - 1))
        target = y[zc_start:zc_end]
        # calc rms
        rms[i] = np.sqrt(np.mean(np.power(target, 2)))
    return rms


def zero_crossing_rate(
    y: NDArray[floating], win_length: int, hop_length: int
) -> NDArray[floating]:
    zcr = np.zeros(int(np.ceil(float(len(y)) / hop_length)))
    for i in range(len(zcr)):
        # get target array
        idx = i * hop_length
        zcr_start = int(max(0, idx - (win_length / 2)))
        zcr_end = int(min(idx + (win_length / 2), len(y) - 1))
        target = y[zcr_start:zcr_end]
        # calc zcr
        sign_arr = np.sign(target)[target != 0 & ~np.isnan(target)]
        zcr[i] = np.sum(np.abs(np.diff(sign_arr)) != 0) / len(target)
    return zcr


def normalize(y: NDArray[floating]) -> NDArray[floating]:
    return (y - float(y.min())) / (float(y.max()) - float(y.min()))


def slide_index(
    goto_min: bool,
    y: NDArray[floating],
    start_idx: int,
    threshold: float,
    margin: int,
) -> int:

    stop_idx: int = -1 if goto_min else len(y)
    step: int = -1 if goto_min else 1

    for i in range(start_idx, stop_idx, step):
        if threshold <= y[i]:
            a_check_end = max(0, i - margin) if goto_min else min(i + margin, len(y))
            a_check = y[a_check_end:i] if goto_min else y[i:a_check_end]
            indices_below_threshold = [j for j, b in enumerate(a_check) if b < threshold]
            if indices_below_threshold:  # is not empty
                i = (
                    min(indices_below_threshold)
                    if goto_min
                    else max(indices_below_threshold)
                )
            else:  # indices_below_threshold is empty -> finish!!!
                return i
    return 0 if goto_min else len(y)



from typing import cast

import numpy as np
from numpy import floating
from numpy.typing import NDArray
from scipy.stats import truncnorm


def white(length: int, seed: int) -> NDArray[floating]:
    return cast(
        NDArray[floating],
        truncnorm.rvs(-3, 3, loc=0, scale=1, size=length, random_state=seed),
    )


def blue(length: int, sr: float, seed: int) -> NDArray[floating]:
    offset = int(length / 2)
    # white noise
    wh = white(length + (offset * 2), seed)
    # fft
    WH_f = np.fft.rfft(wh)
    freqs = np.fft.rfftfreq(len(wh), 1 / sr)
    # white -> blue
    BL_f = WH_f * np.sqrt(freqs)
    # irfft
    bl = np.fft.irfft(BL_f)
    # normalize
    bl /= np.abs(bl).max()

    return bl[offset : length + offset]


def pink(length: int, sr: float, seed: int) -> NDArray[floating]:
    offset = int(length / 2)
    # white noise
    wh = white(length + (offset * 2), seed)
    # fft
    WH_f = np.fft.rfft(wh)
    freqs = np.fft.rfftfreq(len(wh), 1 / sr)
    # white -> pink
    mask = freqs > 20.0
    PK_f = np.zeros_like(WH_f)
    PK_f[mask] = WH_f[mask] / np.sqrt(freqs[mask])
    # irfft
    pk = np.fft.irfft(PK_f)
    # normalize
    pk /= np.abs(pk).max()

    return pk[offset : length + offset]
